//Implementation of a passiv MemPool
//Memory has never to be freed. In contrast to MemPoolP any data can be stored anywhere in allocated blocks.
//However, the user has to tell the class from time to time which memory is still used.


#ifndef MEMPOOLP_HPP_
#define MEMPOOLP_HPP_
#define mempoolp_hpp_version	  "$Revision: 1.5 $"

#include "Assertion.hpp"
#include "DbgAlloc.hpp"
#include "Logging.hpp"
#include "BitVector.hpp"
#include "Ptr32.hpp"

//can be used from outside for garbage collection
// it is assumed that for an area of used memory
// exactly the first and the last block/int will be marked (no marks in between)
class FeedbackForm{
	BitVector<Ptr32> borders;
public:
	bool isBorder(const Ptr32& p){return borders.isSet(p);}
	void markBorder(const Ptr32& p){borders.setFlag(p);}
	friend class MemPoolP;
};


class MemPoolP{
	//actual memory where user data is stored
	uint32_t* 	mem;
	uint32_t	mem_end;
	uint32_t	mem_sz;
	//uint32_t	mem_freed_at_gc;

	//heads of linked lists to freed blocks of size > 1
	vector<uint32_t>	freed_blks; //an entry at index i is a head of a list of blocks with size >= 2^(i+1) and < 2^(i+2)

	MemPoolP(const MemPoolP& m);
	MemPoolP& operator=(const MemPoolP& m);


public:
	inline void hardClear();
	MemPoolP():mem(NULL),mem_end(0),mem_sz(0){}//,mem_freed_at_gc(0){}
	~MemPoolP(){hardClear();}
	uint64_t getMemTotal() const{
		return (mem_sz + freed_blks.size()) << 2;}

	void registerFeedbackForm(FeedbackForm& form){asrt(form.borders.size()==0);
		form.borders.ensureSize(mem_sz,false);}
	inline void submitFeedback(FeedbackForm& form);

	inline bool	inRange(const Ptr32& p) const{return p.idx() < mem_sz;}

	inline void increaseTo(const uint64_t req, const uint64_t min_sz);
	void reserveFor(const uint32_t sz){LOG3("Reserve mempool for "<<sz );
		if(sz<=mem_end) return; increaseTo(sz,sz);}

	template<class C>
	C* getRealPtr(const Ptr32& p)	{asrt(!p.isNull()); asrt(p.id<mem_sz);
									 return (C*)&mem[p.id];}

	template<class C>
	C& getRef(const Ptr32& p)		{return (C&)*(&mem[p.id]);}

	void copy(const Ptr32& from, const Ptr32& to){asrt(!from.isNull()); asrt(!to.isNull()); asrt(from.id<mem_sz); asrt(to.id<mem_sz);
		mem[to.id]=mem[from.id];}

	inline Ptr32 alloc(const uint32_t sz);
	inline void  addToFreedList(const uint32_t blk_hd, const uint32_t size);
	inline void	 freeRange(uint32_t blk_hd, const uint32_t blk_end);

private:

	uint32_t sizeToFloorIndex(uint32_t v){ asrt(v > 1);
		uint32_t l = 0; v>>=2;
		while(v){v >>= 1;++l;}
		return l;
	}

	uint32_t sizeToCeilIndex(uint32_t v){ asrt(v > 1);
		uint32_t l = 0; v = (v-1) >> 1;
		while(v){v >>= 1;++l;}
		return l;
	}

	static const uint32_t ALLIGN_MASK 		= ~(uint32_t)3;
	//static const uint32_t ALLIGN_MASK_SML 	= ~(uint32_t)1;
	/*uint32_t allignSizeUp(uint32_t v){
		return v < 3 ? 2 : ((v+3) & ALLIGN_MASK);}

	uint32_t allignSizeDown(uint32_t v){ asrt(v>1);
		return v < 4 ? 2 : (v & ALLIGN_MASK);}*/



};

void	MemPoolP::freeRange(uint32_t blk_hd, const uint32_t blk_end){
	if(blk_end>=mem_sz){
		mem_sz=blk_hd;
		return;
	}

	blk_hd += (blk_hd&1); asrt((blk_hd&1)==0);
	if(blk_hd+2 > blk_end) return;

	if(blk_hd&3){ //not properly aligned -> can only be used for size 2 blocks
		addToFreedList(blk_hd,2);
		blk_hd += 2;
		if(blk_hd+2 > blk_end) return;
	}

	asrt((blk_hd&3)==0); asrt(blk_hd+2<=blk_end);
	addToFreedList(blk_hd,blk_end-blk_hd);
}



void 	MemPoolP::addToFreedList(const uint32_t blk_hd, const uint32_t size){
	asrt(size > 1); asrt((blk_hd&1) ==0); asrt(size==2 || (blk_hd&3)==0);

	uint32_t lidx			= sizeToFloorIndex(size);

	if(lidx >= freed_blks.size()){
		freed_blks.resize(lidx+1,max32);}

	mem[blk_hd]				= freed_blks[lidx];
	mem[blk_hd+1]			= size;
	freed_blks[lidx]		= blk_hd;
}

void MemPoolP::submitFeedback(FeedbackForm& form){
	freed_blks		.clear(); //forget all fragments
	//mem_freed_at_gc = 0;

	asrt(form.borders.size()==mem_sz);
	uint32_t fr_strt = 0; uint32_t usd_strt = 0;

	do{

		//find beginning of used data
		for(usd_strt=fr_strt;usd_strt<mem_sz && !form.borders.isSet(Ptr32(usd_strt));++usd_strt){}

		freeRange(fr_strt,usd_strt);
		//mem_freed_at_gc += usd_strt - fr_strt;

		for(fr_strt=usd_strt+1;fr_strt<mem_sz && !form.borders.isSet(Ptr32(fr_strt));++fr_strt){}
		asrt(fr_strt<mem_sz || usd_strt+1>=mem_sz);

	}while(++fr_strt<mem_sz);

	form.borders.hardClear();
}




Ptr32 MemPoolP::alloc(const uint32_t sz){
	asrt(sz>=2); LOG3("Start allocating memory mem_sz is " << mem_sz << " mem_end is " << mem_end );
	uint32_t hd;

	const uint32_t ceil		= sizeToCeilIndex(sz);
	for(uint32_t lidx = ceil ;lidx < freed_blks.size();++lidx){
		hd					= freed_blks[lidx];
		if(hd == max32) 	continue;//no list with this size

		//found block but may be too large
		asrt(hd+1<mem_sz); asrt((hd&1)==0); asrt(sz==2 || (hd&3)==0);
		uint32_t real_sz	= mem[hd+1]; asrt(real_sz>=sz);
		freed_blks[lidx]	= mem[hd]; //next in list

		freeRange(hd+sz,hd+real_sz);

		return Ptr32(hd);
	}

	uint64_t req 	= (uint64_t)mem_sz + (uint64_t)sz + 3;
	if(req >(uint64_t)mem_end){
		LOG2("Extend mem with size " << mem_sz );
		increaseTo(req+(req>>1),req);
	}

	hd		= mem_sz + (mem_sz&1); asrt((hd&1)==0);
	if((sz!=2) && (hd&3)){
		addToFreedList(hd,2);
		hd += 2; asrt((hd&3)==0);
	}
	mem_sz 	= hd + sz; asrt(mem_sz <= mem_end);
	return Ptr32(hd);
}

void MemPoolP::hardClear(){
	////LOG_ALWAYS("hardClear memsize is " << mem_sz << " mem_end is " << mem_end << " mem pointer is " << (uint64_t)mem << " id " << ID);
	if(mem != NULL){asrt(mem_end != 0);
		free(mem);
		mem			= NULL;
		mem_end		= 0;
		mem_sz		= 0;
		freed_blks	.clear();
	}
}

void MemPoolP::increaseTo(uint64_t req, uint64_t min){
	////LOG_ALWAYS("increaseTo memsize is " << mem_sz << " mem_end is " << mem_end << " mem pointer is " << (uint64_t)mem << " id " << ID);

	asrt(req > mem_end); asrt(mem_sz<=mem_end);

	if(req >= Ptr32::MAX_ID){
		req = Ptr32::MAX_ID - 1;
		if(min>req) goto BAD_ALLOC;
	}
	asrt(min<= req);

	while(true){
		uint32_t* nmem = (uint32_t*)malloc(req*sizeof(uint32_t));

		if(nmem != NULL){

			if(mem!=NULL){
				memcpy(nmem,mem,mem_sz* sizeof(uint32_t));
				free(mem);
			}else{asrt(mem_sz==0);}

			//memset(nmem+mem_sz,0,(req-mem_sz)*sizeof(uint32_t)); //dbg_used.resize(mem_end,max32);

			mem_end = req;
			mem		= nmem;
			return;
		}

		if(req<=min) goto BAD_ALLOC;

		req = (min>>1) + (req>>1);
	}

	BAD_ALLOC:;
	LOG0("Out of mem in mempool P at size " << req);
	exit(EXIT_FAILURE);
}




#endif /* MEMPOOLP_HPP_ */
