/*
 * VecOfLists.hpp
 *
 *  Created on: Aug 3, 2010
 *      Author: kottlers
 */

#ifndef VECOFLISTS_HPP_
#define VECOFLISTS_HPP_
#define vecoflists_hpp_version	  "$Revision: 1.3 $"

#include "MemPoolA.hpp"
#include "BitVector.hpp"
#include <vector>
using std::vector;



struct List{
	Ptr32 strt;
	Ptr32 next;
	List(const Ptr32& b=Ptr32(), const Ptr32& n=Ptr32()):strt(b),next(n){};
	void clear(){strt=next=Ptr32();}
	bool empty()const{return strt==next;}
private:
	Ptr32 realStart() const{return strt + (strt!=next);}
	uint32_t realSize()const{asrt(next.idx()>=strt.idx()); return next.idx() - strt.idx();}
	Ptr32 getLast(uint32_t sz)	const {return strt == next ? Ptr32() : next - sz;}

	template<class C>
	friend class Range;

	template<class C>
	friend class WIter;

	template<class C>
	friend class Iter;

	template<class E, class A, class Conv>
	friend class VecOfLists;
};


template<class E>
struct Range{
	Ptr32 		curr;
	Ptr32 		end;
	void inc()				{asrt(curr!=end);curr+=(sizeof(E)/4); }
	bool hasNext()const		{return curr!=end;}
	Range(const List& l):curr(l.realStart()),end(l.next){asrt(curr<=end);}
};


template<class E>
struct Iter : public Range<E>{
	typedef Range<E> sup;
	const MemPoolA&	mem;
	Iter(const MemPoolA& pool, const List& l):sup::Range(l),mem(pool){}
	bool getNext(E& v){if(sup::curr==sup::end)return false;
							v=mem.getRef<E>(sup::curr);sup::inc();return true;}
	void skipNext(){sup::inc();}
};

template<class E>
struct WIter: public Range<E>{
	typedef Range<E> sup;
	MemPoolA&	mem;
	Ptr32 		cp;
	WIter(MemPoolA& pool, const List& l):sup::Range(l),mem(pool),cp(sup::curr){}
	bool getNext(E& v)	{if(sup::curr==sup::end)return false;
						v=mem.getRef<E>(sup::curr);sup::inc();return true;}
	//void keepNext()		{asrt(sizeof(E)==4);asrt(Range<E>::hasNext());
	//						mem.copy(sup::curr,cp);cp+=(sizeof(E)/4);sup::inc();}
	void keep(const E& v){asrt(cp<sup::end);mem.getRef<E>(cp)=v;cp+=(sizeof(E)/4);}
};

template<class E, class A, class Conv = DfltKeyTypeConverter<A> >
class VecOfLists{
	vector<List>		lists;
	MemPoolA			mem;
	Conv			conv;

	Ptr32&			getEnd(List& l){asrt(mem.validEndData(l.strt)); return mem.getRef<Ptr32>(l.strt);}
	inline void		ensureMem(List& l, const uint32_t& sz);
	inline void		allocMem(List& l,uint32_t sz);
	inline void		finish(const A& a, const Ptr32& lst, const Ptr32& end);

public:
	VecOfLists(const char w):mem(w){}
	uint64_t getMemTotal() const	{return mem.getMemTotal() + lists.size()*sizeof(List);}
	void reserveForTotal(uint32_t tot){mem.reserveFor(tot);}
	uint32_t listCount() const		{return lists.size();}
	bool empty(const A& a) const	{asrt(conv.toInt(a)<lists.size());return lists[conv.toInt(a)].empty();}
	void enlarge(const uint32_t& sz){asrt(sz>lists.size());//lists.reserve(sz); bei reserve wird nicht unbedingt die capacity verdoppelt
									lists.resize(sz);}
	A newEmptyList()				{lists.push_back(List()); return conv.fromInt(lists.size()-1);}
	//A newList(const uint32_t& sz)	{A a=newEmptyList(); ensureMem(lists[conv.toInt(a)],sz+1);return a;}

	WIter<E> getWIter(const A& a)	{asrt(conv.toInt(a)<lists.size());return WIter<E>(mem,lists[conv.toInt(a)]);}
	Iter<E> getIter(const A& a)const{asrt(conv.toInt(a)<lists.size());return Iter<E>(mem,lists[conv.toInt(a)]);}
	Range<E> getRange(const A& a)const{asrt(conv.toInt(a)<lists.size());return Range<E>(lists[conv.toInt(a)]);}
	inline E lastOf(const A& a) const{asrt(conv.toInt(a)<lists.size()); Ptr32 el = lists[conv.toInt(a)].getLast(sizeof(E)/4);
									return el.isNull() ? E() : mem.getRef<E>(el); }

	inline bool getNext(Range<E>& r,E& v){if(!r.hasNext())return false;
									v=mem.getRef<E>(r.curr);r.inc();return true;}
	inline void writeNext(Range<E>& r,const E& v){
									asrt(r.hasNext());mem.getRef<E>(r.curr)=v;r.inc();}
	inline void finishWrite(const A& a, Range<E>& r){finish(a,r.curr,r.end);}
	inline void finish(const A& a, const WIter<E>& iter){finish(a,iter.cp,iter.end);}
	inline void add(const A& a, const E& e);
	inline void clear(const A& a);
	inline uint32_t approxSizeOf(const A& a);
	inline uint32_t sizeOf(const A& a) const;
	inline void hardClear();
};


template<class E, class A, class Conv>
void VecOfLists<E,A,Conv>::finish(const A& a, const Ptr32& lst, const Ptr32& end){asrt(lst<=end);
	if(lst==end) 		return; //size is the same
	List& l	= lists[conv.toInt(a)];	asrt(lst < l.next);
	if(lst == l.strt+1){
		mem.dealloc(l.strt,getEnd(l));
		l.clear();
	}else{
		l.next	= lst;
	}
}


template<class E, class A, class Conv>
void VecOfLists<E,A,Conv>::clear(const A& a){asrt(conv.toInt(a)<lists.size());
	List& l	= lists[conv.toInt(a)];
	if(l.strt.isNull()) return ;
	mem		.dealloc(l.strt,getEnd(l));
	l		.clear();
}

template<class E, class A, class Conv>
void VecOfLists<E,A,Conv>::hardClear(){
	mem		.hardClear();
	lists	.clear();
}

template<class E, class A, class Conv>
uint32_t VecOfLists<E,A,Conv>::sizeOf(const A& a) const {asrt(conv.toInt(a)<lists.size());
	const List& l	= lists[conv.toInt(a)]; asrt(sizeof(E)>=4);
	return (l.realSize() - (!l.empty())) * (sizeof(E)/4);
}

template<class E, class A, class Conv>
uint32_t VecOfLists<E,A,Conv>::approxSizeOf(const A& a){asrt(conv.toInt(a)<lists.size());
	return lists[conv.toInt(a)].realSize() * sizeof(E)/4;
}

template<class E, class A, class Conv>
void VecOfLists<E,A,Conv>::add(const A& a, const E& e){ asrt(conv.toInt(a)<lists.size());
	const uint32_t sz 		= sizeof(E) / 4;
	List& l					= lists[conv.toInt(a)];
	ensureMem(l,sz);
	mem.getRef<E>(l.next)	= e;
	l.next					= l.next + sz; asrt(l.next <= getEnd(l));
}


template<class E, class A, class Conv>
void VecOfLists<E,A,Conv>::allocMem(List& l,uint32_t sz){ asrt(sz>1);
	l.strt 		= mem.alloc(sz);
	l.next 		= l.strt+1;
	getEnd(l) 	= l.strt + sz;
	asrt(mem.validEndData(l.strt)); 	asrt(mem.validEndData(l.strt + (sz-1)));
}


template<class E, class A, class Conv>
void VecOfLists<E,A,Conv>::ensureMem(List& l, const uint32_t& sz){ asrt(sz!=0);
	if(l.strt.isNull()){
		allocMem(l,2*(sz+(sz==1))); return;}

	Ptr32 end		= getEnd(l);
	if(l.next + sz <= end){return;}

	uint32_t currm	= end.idx()-l.strt.idx(); asrt(sz<=currm);

	uint32_t ext	= mem.tryAllocAt(end,l.next.idx()+sz-end.idx(),currm);
	if(ext != 0){
		getEnd(l)	= end + ext; return;	}

	List old		= l;
	allocMem(l,2*currm);

	for(Ptr32 i=old.strt+1;i != old.next;++i){mem.copy(i,l.next);++l.next;}

	LOG3("Move list to new location");
	asrt(mem.validEndData(old.strt)); 	asrt(mem.validEndData(old.strt + (currm-1)));
	mem.dealloc(old.strt,currm);
}



#endif /* VECOFLISTS_HPP_ */
